#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module is the entrypoint when a user calls:
$ python -m treedoc [object]
"""

from treedoc.main import CLI_entrypoint

if __name__ == "__main__":
    CLI_entrypoint()
