#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:14:43 2019

@author: tommy
"""

from treedoctestpackage.module import MyClass
from treedoctestpackage.subpackage.subpackagemodule import func_subtraction

MyClass = MyClass
func_subtraction = func_subtraction
